/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy_main_server.hpp
 **
 **  Main server that runs inside the visualised system
 **/

#ifndef _GCSPY_MAIN_SERVER_HPP_

#define _GCSPY_MAIN_SERVER_HPP_

#include <stdio.h>
#include "gcspy_timer.hpp"
#include "gcspy_comm.hpp"
#include "gcspy_space.hpp"
#include "gcspy_server_state.hpp"
#include "gcspy_interpreter.hpp"

#define magic_string "GCspy666"

class gcspyMainServer {
private:
  enum PrivateConstants {
    name_len            = 48,
    max_spaces          = 10,
    general_info_size   = 4096,

    endianess           = 1
  };

  gcspyServer           *_server;
  gcspyClient           *_client;

  gcspy_length_t         _max_len;
  gcspy_bool_t           _verbose;
  gcspy_bool_t           _connected;

  gcspy_bool_t           _shutdown_now;
  gcspy_bool_t           _shutdown_sent;
  gcspy_bool_t           _pause_now;
  gcspy_bool_t           _paused;

  char                   _name[name_len];

  int                    _space_num;
  gcspySpace            *_spaces[max_spaces];

  int                    _event_num;
  char                  *_event_names[gcspyServerState::max_events];

  char                   _general_info[general_info_size];

  gcspy_time_t           _elapsed_time_start;
  gcspy_time_t           _elapsed_timer;
  gcspy_time_t           _compensation_time_start;
  gcspy_time_t           _compensation_timer;

  int                    _event_counts[gcspyServerState::max_events];

  gcspyServerState      *_server_state;

  gcspyInterpreter      *_interpreter;

  inline void verbose_print (const char *str) {
    if (_verbose)
      printf("GCspy Server: %s\n", str);
  }

  inline void start_elapsed_timer () {
    _elapsed_time_start = gcspyTimer::get_time();
  }

  inline void stop_elapsed_timer () {
    gcspy_time_t end = gcspyTimer::get_time();
    gcspy_time_t diff = end - _elapsed_time_start;
    _elapsed_timer = diff;
  }

  inline void reset_compensation_timer () {
    _compensation_timer = (gcspy_time_t) 0;
  }
  
  gcspy_bool_t check_client (void);
  void internal_setup (void);
  void send_boot_info (void);
  gcspy_bool_t internal_loop (void);
  void internal_safepoint (void);

public:
  gcspyMainServer (int port,
		   const char *name,
		   gcspy_length_t max_len = 0,
		   gcspy_bool_t verbose = 0);

  inline void set_general_info (const char *general_info) {
    gcspyUtils::copy_string(_general_info, general_info, general_info_size);
  }

  void add_space (gcspySpace *space);

  inline gcspySpace *get_space (gcspy_index_t index) {
    gcspy_assert(index < _space_num);
    gcspy_assert(_spaces[index] != NULL);
    return _spaces[index];
  }

  void add_event (gcspy_index_t index, const char *event_name);

  gcspy_bool_t connect (void);
  void wait_for_client (void);
  void main_loop (void);

  void transmit_streams (void);

  inline gcspy_bool_t is_connected (void) {
    return _connected;
  }

  inline gcspy_bool_t is_connected (gcspy_index_t event) {
    return ( _connected && _server_state->is_enabled(event) &&
	    ((_event_counts[event] % _server_state->get_period(event)) == 0) );
  }

  inline void safepoint (gcspy_index_t event) {
    gcspy_assert(event < _event_num);
    ++_event_counts[event];
    if (is_connected(event)) {
      stop_elapsed_timer();
      _interpreter->send_event_count_cmd(_event_counts, _event_num);
      gcspy_assert(((gcspy_time_t) ((int) _elapsed_timer)) == _elapsed_timer);
      gcspy_assert(((gcspy_time_t) ((int) _compensation_timer)) ==
		   _compensation_timer);
      _interpreter->send_event_cmd(event, _elapsed_timer, _compensation_timer);

      if (_server_state->get_delay(event))
	gcspyTimer::wait(_server_state->get_delay(event));
      if (_server_state->is_paused(event))
	_pause_now = 1;
      internal_safepoint();
      reset_compensation_timer();
      start_elapsed_timer();
    }
  }

  inline void safepoint () {
    if (is_connected()) {
      internal_safepoint();
    }
  }

  inline void start_compensation_timer () {
    _compensation_time_start = gcspyTimer::get_time();
  }

  inline void stop_compensation_timer () {
    gcspy_time_t end = gcspyTimer::get_time();
    gcspy_time_t diff = end - _compensation_time_start;
    _compensation_timer += diff;
  }
};

#endif //_GCSPY_MAIN_SERVER_HPP_
