/**
 ** $Id: Interpreter.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.interpreter;

import gcspy.comm.BufferedInput;
import gcspy.comm.BufferedOutput;
import gcspy.comm.CommandStream;
import gcspy.comm.SocketClient;

import java.io.IOException;

/**
 * Facilities shared by the client and server interpreters
 * 
 * @author Tony Printezis
 */
abstract public class Interpreter extends CommandStream {

	// Constants
	static private final String MAGIC_STRING = "GCspy666";
	static private final int ENDIANESS = 1;
	static protected final int DEFAULT_MAX_LEN = (4 * 1024 * 1024);

	// Shared data
	/** The name of the interpreter */
	protected String name;
	/** The client */
	protected SocketClient client;

	private TraceData data;
	
	protected TraceData getData() {
		return this.data;
	}

	// Commands
	/** Client has requested pause command */
	static protected final byte PAUSE_REQ_CMD = FIRST_AVAILABLE_CMD;
	/** Server has paused command */
	static protected final byte PAUSE_CMD = FIRST_AVAILABLE_CMD + 1;
	/** Client has requested restart command */
	static protected final byte RESTART_CMD = FIRST_AVAILABLE_CMD + 2;
	/** Client has requested play one event command */
	static protected final byte PLAY_ONE_CMD = FIRST_AVAILABLE_CMD + 3;
	/** Client has requested shutdown command */
	static protected final byte SHUTDOWN_REQ_CMD = FIRST_AVAILABLE_CMD + 4;
	/** Server has shut down command */
	static protected final byte SHUTDOWN_CMD = FIRST_AVAILABLE_CMD + 5;
	/** Start new stream command */
	static protected final byte STREAM_CMD = FIRST_AVAILABLE_CMD + 6;
	/** Event occurred command */
	static protected final byte EVENT_CMD = FIRST_AVAILABLE_CMD + 7;
	/** Start new control stream command */
	static protected final byte CONTROL_CMD = FIRST_AVAILABLE_CMD + 8;
	/** Event filters command */
	static protected final byte EVENT_FILTERS_CMD = FIRST_AVAILABLE_CMD + 9;
	/** Event count command */
	static protected final byte EVENT_COUNT_CMD = FIRST_AVAILABLE_CMD + 10;
	/** Summary information command */
	static protected final byte SUMMARY_CMD = FIRST_AVAILABLE_CMD + 11;
	/** Space information command */
	static protected final byte SPACE_INFO_CMD = FIRST_AVAILABLE_CMD + 12;
	/** Space command */
	static protected final byte SPACE_CMD = FIRST_AVAILABLE_CMD + 13;

	static private final byte CMD_LEN = SPACE_CMD + 1;

	/******************** Inner Classes ********************/

	protected class BootstrapParameters {
		public boolean pauseAtStart;

		public BootstrapParameters() {
		}

		public BootstrapParameters(boolean pauseAtStart) {
			this.pauseAtStart = pauseAtStart;
		}
	}

	/******************** Accessor Methods ********************/

	/**
	 * Get the interpreter's name
	 * 
	 * @return the name of the interpreter
	 */
	public String getName() {
		return name;
	}

	/******************** Utilities ********************/

	/**
	 * Send a single command
	 * 
	 * @param cmd
	 *            the command to send
	 * @throws IOException
	 */
	protected void sendSingleCommand(byte cmd) throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, cmd);
		finish(output);

		client.send(output);
	}

	/******************** Boot Info ********************/

	/**
	 * Send boot information
	 * 
	 * @param server
	 *            True if this is the server
	 * @param params
	 *            the BootStrapParameters
	 * @throws IOException
	 */
	protected void sendBootInfo(boolean server, BootstrapParameters params)
			throws IOException {
		BufferedOutput output = client.createBufferedOutput();
		output.writeString(MAGIC_STRING);
		output.writeInt(ENDIANESS);

		if (server) {
			output.writeString(name);
		} else {
			output.writeBoolean(params.pauseAtStart);
		}
		client.send(output);
	}

	/**
	 * Receive boot information
	 * 
	 * @param server
	 *            True if this is the server
	 * @param params
	 *            the BootStrapParameters
	 * @throws IOException
	 * @throws InterpreterException
	 *             if magic string or endianness do not match
	 */
	protected void receiveBootInfo(boolean server, BootstrapParameters params)
			throws IOException {
		client.definitelyReceive();
		BufferedInput input = client.createBufferedInput();
		if (!input.readString().equals(MAGIC_STRING))
			throw new InterpreterException("Magic String does not match");
		if (input.readInt() != ENDIANESS)
			throw new InterpreterException("Endianness field does not match");

		if (!server) {
			name = input.readString();
		} else {
			params.pauseAtStart = input.readBoolean();
		}
		input.close();
	}

	/******************** Debugging ********************/

	public void dump() {
		System.out.println("== Interpreter");
		System.out.println("--   Name: " + getName());
		System.out.println("--   " + data.getSpaceNum() + " space(s)");
		data.getEvents().dump();
	}

	public void dumpSpaces() {
		for (int i = 0; i < data.getSpaceNum(); ++i)
			data.getSpace(i).dump();
	}

	/******************** Constructors ********************/

	/**
	 * Setup the event filters for all events
	 */
	protected void setupEventFilters() {
		data.setEventFilters(null);
		data.setEventFilters(new EventFilters(data.getEvents().getNum()));
	}

	/**
	 * Setup counts of events for all events
	 */
	protected void setupEventCount() {
		data.setEventCount(new int[data.getEvents().getNum()]);
		for (int i = 0; i < data.getEventCount().length; ++i)
			data.getEventCount()[i] = 0;
	}

	/** Create an interpreter */
	public Interpreter(TraceData data) {
		super(CMD_LEN);
		this.data = data;
	}

	/**
	 * Create an interpreter
	 * 
	 * @param name
	 *            The name of the interpreter
	 */
	public Interpreter(String name, TraceData data) {
		this(data);
		this.name = name;
	}

}
