/**
 ** $Id: ClientInterpreter.java 21 2005-06-11 00:25:23Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.interpreter.client;

import gcspy.comm.BufferedInput;
import gcspy.comm.BufferedOutput;
import gcspy.comm.Command;
import gcspy.comm.SocketClient;
import gcspy.interpreter.EventFilters;
import gcspy.interpreter.InputGenerator;
import gcspy.interpreter.Interpreter;
import gcspy.interpreter.InterpreterException;
import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.StreamData;
import gcspy.interpreter.TraceData;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * The client interpreter
 * 
 * @author Tony Printezis
 * @author <a href="http://www/cs/kent.ac.uk">Richard Jones</a>
 */
public class ClientInterpreter extends Interpreter {
	
	private static final Logger logger = Logger.getLogger(ClientInterpreter.class.getName());

	private boolean terminated = false;

	// Listeners
	private List<PauseListener> pauseListeners = new ArrayList<PauseListener>();

	private List<EventCommandListener> eventListeners = new ArrayList<EventCommandListener>();

	private List<SpaceCommandListener> spaceListeners = new ArrayList<SpaceCommandListener>();
	
	private List<ConnectListener> connectListeners = new ArrayList<ConnectListener>();
	
	/** ****************** Accessor Methods ******************* */

	/**
	 * Get a client space
	 * 
	 * @param i
	 *            the space's ID
	 * @return The client space
	 */
	public Space getSpace(int i) {
		return getData().getSpace(i);
	}

	/** ****************** Listeners ******************* */

	// Pause

	/**
	 * Add a pause listener
	 * 
	 * @param pauseListener
	 *            The listener
	 */
	public void addPauseListener(PauseListener pauseListener) {
		pauseListeners.add(pauseListener);
	}

	/**
	 * Tell all pause listeners to pause
	 */
	public void firePauseListeners() {
		for (PauseListener listener : pauseListeners)
			listener.onPause();
	}
	
	public void addConnectListener(ConnectListener commandListener) {
		connectListeners.add(commandListener);
	}
	
	public void fireConnectListeners() {
		for (ConnectListener l : connectListeners) {
			l.onConnect();
		}
	}

	// Event

	/*
	 * we need synchronization in the next three methods because the add/remove
	 * calls modify the number of elements in the list, and the iterator further
	 * down might throw a concurrent modification exception
	 */

	/**
	 * Add an event listener. We need synchronization because the add call
	 * modifies the number of elements in the list, and iterators might throw a
	 * concurrent modification exception.
	 * 
	 * @param eventListener
	 *            the event listener
	 */
	synchronized public void addEventListener(EventCommandListener eventListener) {
		eventListeners.add(eventListener);
	}

	/**
	 * Remove an event listener We need synchronization because the remove call
	 * modifies the number of elements in the list, and iterators might throw a
	 * concurrent modification exception.
	 * 
	 * @param eventListener
	 *            the event listener
	 */
	synchronized public void removeEventListener(EventCommandListener eventListener) {
		eventListeners.remove(eventListener);
	}

	/**
	 * Call the event listeners for an event We need synchronization because the
	 * add/remove call modifies the number of elements in the list, and the
	 * iterator might throw a concurrent modification exception.
	 * 
	 * @param eventID
	 *            The event's ID
	 * @param elapsedTime
	 *            The elapsed time
	 * @param compensationTime
	 *            The compensation time
	 */
	synchronized public void fireEventListeners(int eventID, int elapsedTime,
			int compensationTime) {
		for (EventCommandListener listener : eventListeners)
			listener.onEventCmd(eventID, elapsedTime, compensationTime);
	}

	// Space

	/**
	 * Add a space listener
	 * 
	 * @param spaceListener
	 *            the listener.
	 */
	public void addSpaceListener(SpaceCommandListener spaceListener) {
		spaceListeners.add(spaceListener);
	}

	/**
	 * Remove a space listener
	 * 
	 * @param spaceListener
	 *            the listener.
	 */
	public void removeSpaceListener(EventCommandListener spaceListener) {
		spaceListeners.remove(spaceListener);
	}

	/**
	 * Call the space listeners for a space
	 * 
	 * @param space
	 *            The space
	 */
	private void fireSpaceListeners(Space space) {
		for (SpaceCommandListener listener : spaceListeners)
			listener.onSpaceCmd(space);
	}

	/** ****************** Commands ******************* */

	/** *** PAUSE REQ **** */

	private class PauseReqCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("PAUSE REQ Command not implemented");
		}
	}

	/**
	 * Send a pause request command
	 * 
	 * @throws IOException
	 */
	public void sendPauseReq() throws IOException {
		sendSingleCommand(PAUSE_REQ_CMD);
	}

	/** *** PAUSE **** */

	private class PauseCmd implements Command {
		public void execute(BufferedInput input) {
			firePauseListeners();
		}
	}

	/** *** RESTART **** */

	private class RestartCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("RESTART Command not implemented");
		}
	}

	/**
	 * Send a restart command
	 * 
	 * @throws IOException
	 */
	public void sendRestart() throws IOException {
		sendSingleCommand(RESTART_CMD);
	}

	/** *** PLAY ONE **** */

	private class PlayOneCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("PLAY ONE Command not implemented");
		}
	}

	/**
	 * Send a play one command
	 * 
	 * @throws IOException
	 */
	public void sendPlayOne() throws IOException {
		sendSingleCommand(PLAY_ONE_CMD);
	}

	/** *** SHUTDOWN REQ **** */

	private class ShutdownReqCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException(
					"SHUTDOWN REQ Command not implemented");
		}
	}

	/**
	 * Send a shut down request command
	 * 
	 * @throws IOException
	 */
	public void sendShutdownReq() throws IOException {
		sendSingleCommand(SHUTDOWN_REQ_CMD);
	}

	/** *** SHUTDOWN **** */

	private class ShutdownCmd implements Command {
		public void execute(BufferedInput input) {
			terminated = true;
		}
	}

	/** *** STREAM **** */

	private class StreamCmd implements Command {
		public void execute(BufferedInput input) {

			int spaceID = (int) input.readByte();
			int streamID = (int) input.readByte();

			Space space = getSpace(spaceID);
			Stream stream = space.getStream(streamID);
		
			switch (stream.getDataType()) {
			case Byte:
				stream.setData(new StreamData(input.readByteArray()));
				break;
			case Short:
				stream.setData(new StreamData(input.readShortArray()));
				break;
			case Int:
				stream.setData(new StreamData(input.readIntArray()));
				break;
			}
		}
	}

	/** *** EVENT **** */

	private class EventCmd implements Command {
		public void execute(BufferedInput input) {
			int eventID = (int) input.readByte();
			int elapsedTime = input.readInt();
			int compensationTime = input.readInt();
			fireEventListeners(eventID, elapsedTime, compensationTime);
		}
	}

	/** *** CONTROL **** */

	private class ControlCmd implements Command {
		public void execute(BufferedInput input) {
			int spaceID = (int) input.readByte();

			byte control[] = input.readByteArray();

			getSpace(spaceID).setControl(control);
		}
	}

	/** *** EVENT FILTERS **** */

	private class EventFiltersCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException(
					"EVENT FILTERS Command not implemented");
		}
	}

	/**
	 * Send the client's event filters
	 * 
	 * @throws IOException
	 */
	public void sendEventFilters() throws IOException {
		sendEventFilters(getData().getEventFilters());
	}

	/**
	 * Send the event filters
	 * 
	 * @param eventFilters
	 *            The event filters to send
	 * @throws IOException
	 */
	public void sendEventFilters(EventFilters eventFilters) throws IOException {
		BufferedOutput output = client.createBufferedOutput();
		int len = eventFilters.getNum();

		start(output);
		putCmd(output, EVENT_FILTERS_CMD);
		output.writeShort((short) len);

		boolean enabled[] = eventFilters.getEnabled();
		int delays[] = eventFilters.getDelays();
		boolean pauses[] = eventFilters.getPauses();
		int periods[] = eventFilters.getPeriods();

		for (int i = 0; i < len; ++i) {
			output.writeBoolean(enabled[i]);
			output.writeInt(delays[i]);
			output.writeBoolean(pauses[i]);
			output.writeInt(periods[i]);
		}
		finish(output);

		client.send(output);
	}

	/** *** EVENT COUNT **** */

	private class EventCountCmd implements Command {
		public void execute(BufferedInput input) {
			int count[] = input.readIntArray();
			getData().setEventCount(count);
		}
	}

	/** *** SUMMARY **** */

	private class SummaryCmd implements Command {
		public void execute(BufferedInput input) {
			int spaceID = (int) input.readByte();
			int streamID = (int) input.readByte();
			int summary[] = input.readIntArray();

			getSpace(spaceID).getStream(streamID).setSummary(summary);
		}
	}

	/** *** SPACE INFO **** */

	private class SpaceInfoCmd implements Command {
		public void execute(BufferedInput input) {
			int spaceID = (int) input.readByte();
			String spaceInfo = input.readString();

			getSpace(spaceID).setSpaceInfo(spaceInfo);
		}
	}

	/** *** SPACE **** */

	private class SpaceCmd implements Command {
		public void execute(BufferedInput input) {
			Space space = Space.deserialise(input);
			getData().setSpace(space.getID(), space);
			fireSpaceListeners(space);
		}
	}

	/** ****************** Configuration Communication ******************* */

	private class ClientInputGenerator implements InputGenerator {

		private SocketClient client;

		public BufferedInput createBufferedInput() throws IOException {
			client.definitelyReceive();
			return client.createBufferedInput();
		}

		ClientInputGenerator(SocketClient client) {
			this.client = client;
		}
	}

	private void deserialiseEverything() throws IOException {
		getData().deserialiseEverything(new ClientInputGenerator(client));
	}

	/** ****************** Connection ******************* */

	private void checkServer(BootstrapParameters params) throws IOException {
		/* client sends first */
		sendBootInfo(false, params);
		receiveBootInfo(false, null);
	}

	private void setupClient(String server, int port, int maxLen)
			throws IOException {
		/*
		 * make sure we clear up the client variable before we re-assign it,
		 * otherwise we need double the memory requirements
		 */
		client = null;
		client = new SocketClient(server, port, maxLen);
	}

	/**
	 * Connect to the server
	 * 
	 * @param server
	 *            The server's host name
	 * @param port
	 *            The port on which to connect
	 * @param pauseAtStart
	 *            Whether the server should pause at the start
	 * @throws IOException
	 */
	public void connectToServer(String server, int port, boolean pauseAtStart)
			throws IOException {
		connectToServer(server, port, pauseAtStart, DEFAULT_MAX_LEN);
	}

	/**
	 * Connect to the server
	 * 
	 * @param server
	 *            The server's host name
	 * @param port
	 *            The port on which to connect
	 * @param pauseAtStart
	 *            Whether the server should pause at the start
	 * @param maxLen
	 *            The maximum length of the communication buffers
	 * @throws IOException
	 */
	public void connectToServer(String server, int port, boolean pauseAtStart,
			int maxLen) throws IOException {
		if (logger.isLoggable(Level.INFO)) {
			logger.log(Level.INFO, "connecting to server " + server + ", port " + port);
		}
		BootstrapParameters params = new BootstrapParameters(pauseAtStart);
		setupClient(server, port, maxLen);
		logger.log(Level.INFO, "connected to server");
		checkServer(params);
		deserialiseEverything();
		setupEventFilters();
		
		fireConnectListeners();
		
		// notify about initial space changes
		for (int i = 0; i < getData().getSpaceNum(); i++) {
			fireSpaceListeners(getData().getSpace(i));
		}
	}

	/** ****************** Main Loop ******************* */

	/**
	 * The main client interpreter loop. It repeatedly receives commands and
	 * executes them
	 */
	public void mainLoop() throws IOException {
		logger.log(Level.FINE, "starting client main loop");
		while (true) {
			client.receive();
			if (client.hasTerminated()) {
				logger.log(Level.WARNING, "main loop terminated (server died)");
				break;
			}
			terminated = false;
			execute(client.createBufferedInput());
			if (terminated) {
				logger.log(Level.INFO, "main loop terminated (shutdown received)");
				break;
			}
		}
		client.close();
	}
	
	/** ****************** Constructors ******************* */

	/** Create a new ClientInterpreter */

	public ClientInterpreter(TraceData data) {
		super(data);
		init();
	}
	
	private void init() {
		cmds[PAUSE_REQ_CMD] = new PauseReqCmd();
		cmds[PAUSE_CMD] = new PauseCmd();
		cmds[RESTART_CMD] = new RestartCmd();
		cmds[PLAY_ONE_CMD] = new PlayOneCmd();
		cmds[SHUTDOWN_REQ_CMD] = new ShutdownReqCmd();
		cmds[SHUTDOWN_CMD] = new ShutdownCmd();
		cmds[STREAM_CMD] = new StreamCmd();
		cmds[EVENT_CMD] = new EventCmd();
		cmds[CONTROL_CMD] = new ControlCmd();
		cmds[EVENT_FILTERS_CMD] = new EventFiltersCmd();
		cmds[EVENT_COUNT_CMD] = new EventCountCmd();
		cmds[SUMMARY_CMD] = new SummaryCmd();
		cmds[SPACE_INFO_CMD] = new SpaceInfoCmd();
		cmds[SPACE_CMD] = new SpaceCmd();				
	}

}
