package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.TraceData;
import gcspy.interpreter.client.ClientInterpreter;
import gcspy.interpreter.client.ConnectListener;
import gcspy.tools.EventStoredListener;
import gcspy.tools.OutputStreamTrace;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * {@link OutputStreamTrace} that keeps an index with offset and sizes of events
 * and spaces.
 * 
 * @author Thomas Schatzl
 * @author Peter Hofer
 */
public class Indexer extends OutputStreamTrace {

	private static final Logger logger = Logger.getLogger(Indexer.class.getName());

	private static class IndexData {
		public long offset;
		public long size;

		IndexData(long offset, long size) {
			this.offset = offset;
			this.size = size;
		}
	}

	private EventStoredListener eventStoredListener = new EventStoredListener() {
		@Override
		public void spaceStored(Space space, long startPos, long size) {
			if (logger.isLoggable(Level.FINE)) {
				logger.log(Level.FINE, "Space: " + space.getID() + ", bytes written: " + size);
			}

			if ((spaceIndex.size() - 1) < space.getID()) {
				for (int i = 0; i < space.getID() - spaceIndex.size() + 1; i++) {
					spaceIndex.add(null);
				}
			}
			spaceIndex.set(space.getID(), space);
		}

		@Override
		public void eventStored(int event, long startPos, long size) {
			try {
				flush();
			} catch (IOException e) {
				logger.log(Level.WARNING, "Error flushing buffer (ignoring)", e);
			}

			if (logger.isLoggable(Level.FINE)) {
				logger.log(Level.FINE, "Event: " + event + ", bytes written: " + size);
			}

			frameIndex.add(new IndexData(startPos, size));
		}
	};

	private ConnectListener connectListener = new ConnectListener() {
		@Override
		public void onConnect() {
			try {
				getData().serialiseEverything(Indexer.this);
				flush();
			} catch (IOException e) {
				throw new RuntimeException("Must not fail");
			}
		}
	};

	private final List<IndexData> frameIndex = new ArrayList<IndexData>();
	private final List<Space> spaceIndex = new ArrayList<Space>();

	private final ClientInterpreter interpreter;
	private final TraceData interpreterData;

	/**
	 * Constructor.
	 * 
	 * @param interpreter
	 *            Client interpreter for subscribing to events.
	 * @param data
	 *            Trace data.
	 * @param out
	 *            Output stream to write data to.
	 */
	public Indexer(ClientInterpreter interpreter, TraceData data, OutputStream out) {
		super(out, OutputStreamTrace.INFINITE, Compression.None);

		this.interpreter = interpreter;
		this.interpreterData = data;

		interpreter.addEventListener(this);
		interpreter.addSpaceListener(this);
		interpreter.addConnectListener(connectListener);

		addEventStoredListener(eventStoredListener);
	}

	public long getFramePosAt(int index) {
		return frameIndex.get(index).offset;
	}

	public long getFrameSizeAt(int index) {
		return frameIndex.get(index).size;
	}

	public int getNumEvents() {
		return frameIndex.size();
	}

	public int getNumSpaces() {
		return spaceIndex.size();
	}

	public Space getSpace(int index) {
		return spaceIndex.get(index);
	}

	protected TraceData getData() {
		return interpreterData;
	}

	protected ClientInterpreter getInterpreter() {
		return interpreter;
	}
}
