/**
 ** $Id: LegendFrame.java 24 2005-06-17 09:37:56Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.client.EventCommandListener;
import gcspy.interpreter.client.SpaceCommandListener;
import gcspy.vis.TileComponent.TileFlags;
import gcspy.vis.TileComponent.TileType;

import java.util.Arrays;
import java.util.EnumSet;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import net.miginfocom.swing.MigLayout;

/**
 * Component that shows a legend for one or more streams of a space.
 * 
 * @author Tony Printezis
 * @author Peter Hofer
 */
class LegendView extends JComponent {

	private Space space;
	private int[] streamIds;
	private StreamLegend[] streamLegends;
	private TileStyle tileStyle;
	private JPanel specialPanel;
	private TileComponent unusedTile;
	private TileComponent illegalTile;

	private EventCommandListener eventCommandListener = new EventCommandListener() {
		public void onEventCmd(int eventID, int elapsedTime, int compensationTime) {
			int count = 0;
			int[] changed = new int[streamIds.length]; // streams to update
			for (int streamId : streamIds) {
				Stream stream = space.getStream(streamId);
				if (stream.getPresentation() == Stream.PRESENTATION_MAX_VAR) {
					changed[count++] = streamId;
				}
			}
			changed = Arrays.copyOf(changed, count); // truncate
			updateStreams(changed);
		}
	};

	private SpaceCommandListener spaceListener = new SpaceCommandListener() {
		public void onSpaceCmd(Space space) {
			if (LegendView.this.space.getID() == space.getID()) {
				LegendView.this.space = space;
			}
		}
	};

	/**
	 * Update the legends of the streams with the specified ids.
	 * 
	 * @param changed
	 *            Arrays that contains some, all or none of the ids from
	 *            {@link #streamIds} in the same order.
	 */
	private void updateStreams(int[] changed) {
		if (changed.length == 0)
			return;

		removeAll();
		int k = 0; // search index
		for (int i = 0; i < streamIds.length; ++i) {
			while (k < changed.length && changed[k] != streamIds[i]) {
				k++;
			}
			if (k < changed.length) {
				// found in array
				streamLegends[i] = new StreamLegend(space, streamIds[i], tileStyle);
			}
			add(streamLegends[i]);
		}

		revalidate(); // will not repaint otherwise
	}

	/**
	 * Clears and populates this component with {@link StreamLegend} components
	 * for all streams.
	 */
	private void populate() {
		streamLegends = new StreamLegend[streamIds.length];
		updateStreams(streamIds);

		// Panel with special types of tiles (unused, illegal)
		if (specialPanel == null) {
			specialPanel = new JPanel(new MigLayout());

			unusedTile = new TileComponent(tileStyle);
			specialPanel.add(unusedTile, "wmax pref, hmax pref");
			specialPanel.add(new JLabel(space.getUnusedString()), "push, wrap");

			illegalTile = new TileComponent(tileStyle);
			specialPanel.add(illegalTile, "wmax pref, hmax pref");
			specialPanel.add(new JLabel(space.getIllegalString()), "push");
		}
		unusedTile.setTile(space, 0, TileType.UNUSED, EnumSet.noneOf(TileFlags.class), 0);
		illegalTile.setTile(space, 0, TileType.ILLEGAL, EnumSet.noneOf(TileFlags.class), 0);
		add(specialPanel);

		revalidate(); // will not repaint otherwise
	}

	/**
	 * Set the streams displayed in this legend view.
	 * 
	 * @param streamIds
	 *            The stream's ID
	 */
	public void setDisplayedStreams(int[] streamIds) {
		this.streamIds = streamIds;
		update();
	}

	/**
	 * Update the displayed legend (e.g. after a tile style change).
	 */
	public void update() {
		populate();
	}

	/**
	 * Construct a new legend frame
	 * 
	 * @param interpreter
	 *            The client interpreter
	 * @param space
	 *            The space for which stream legends should be displayed.
	 * @param streamIds
	 *            Ids of the streams to display initially.
	 * @param tileStyle
	 *            The tile style configuration to use
	 */
	LegendView(InterpreterEventProvider interpreter, Space space, int[] streamIds, TileStyle tileStyle) {
		this.space = space;
		this.streamIds = streamIds;
		this.tileStyle = tileStyle;

		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		populate();

		interpreter.addEventListener(eventCommandListener);
		interpreter.addSpaceListener(spaceListener);
	}
}
