package gcspy.vis;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.image.BufferedImage;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

/**
 * Panel that displays an area inside its bounds showing the region near the
 * mouse pointer in upscaled form.
 * 
 * <p>
 * In order to capture all mouse events, an additional overlay is used
 * (resembling a <i>glass pane</i>) and after processing the events, they are
 * dispatched to their actual targets in a <i>content pane</i>. The methods
 * {@link #setLayout(LayoutManager)}, {@link #remove(Component)} and all
 * variants of {@link #add(Component)} were overridden to work on the content
 * pane instead of this panel directly (similar to what {@link JFrame} does).
 * </p>
 * 
 * <p>
 * Magnification is achieved by having all child components draw into an image
 * buffer, then painting the contents of that buffer on the screen and then
 * extracting and scaling the region to magnify from the buffer and painting it
 * in the magnifying area on the screen.
 * </p>
 */
public class MagnifyingPanel extends JPanel {

	private static final int DEFAULT_MAGNIFICATION_FACTOR = 2;

	private BufferedImage paintBuffer = null;

	private Point magnifierCornerGap = new Point(5, 5);
	private Dimension magnifierSize = new Dimension(120, 90);
	private int magnificationFactor = DEFAULT_MAGNIFICATION_FACTOR;
	private boolean magnificationEnabled = true;

	private Point magnifiedCenter = null;
	private Rectangle magnifiedArea = null;
	private Rectangle magnifierBounds = null;
	
	private Container contentPane;
	private Container layeredPane;
	private boolean contentPaneForward;

	/**
	 * Creates an empty magnifying panel.
	 */
	public MagnifyingPanel() {
		contentPane = new JPanel();

		layeredPane = new JLayeredPane();
		layeredPane.add(contentPane, Integer.valueOf(0));
		layeredPane.add(glassPane, Integer.valueOf(Integer.MAX_VALUE));

		contentPaneForward = false;
		setLayout(new BorderLayout());
		add(layeredPane, BorderLayout.CENTER);
		// Forward add(), remove() and setLayout() calls to the
		// content pane object from now on
		contentPaneForward = true;

		glassPane.addMouseListener(mouseListener);
		glassPane.addMouseMotionListener(mouseListener);
		glassPane.addMouseWheelListener(mouseListener);
	}
	
	/**
	 * Enables or disables magnification.
	 * 
	 * @param enabled
	 *            <code>true</code> to enable, <code>false</code> to disable.
	 */
	public void setMagnificationEnabled(boolean enabled) {
		magnificationEnabled = enabled;
	}

	/**
	 * @return <code>true</code> when magnification is currently enabled,
	 *         <code>false</code> otherwise.
	 */
	public boolean isMagnificationEnabled() {
		return magnificationEnabled;
	}

	/**
	 * @return the actual container for user interface components.
	 */
	public Container getContentPane() {
		return contentPane;
	}

	@Override
	public void doLayout() {
		super.doLayout();

		// Make glass pane and content pane span the whole panel
		int width = layeredPane.getWidth();
		int height = layeredPane.getHeight();
		contentPane.setBounds(0, 0, width, height);
		glassPane.setBounds(0, 0, width, height);
	}

	@Override
	protected void addImpl(Component comp, Object constraints, int index) {
		if (contentPaneForward) {
			getContentPane().add(comp, constraints, index);
		} else {
			super.addImpl(comp, constraints, index);
		}
	}
	
	@Override
	public void remove(Component c) {
		if (contentPaneForward) {
			getContentPane().remove(c);
		} else {
			super.remove(c);
		}
	}

	@Override
	public void setLayout(LayoutManager mgr) {
		if (contentPaneForward) {
			getContentPane().setLayout(mgr);
		} else {
			super.setLayout(mgr);
		}
	}
	
	private Container glassPane = new JComponent() {
		// Initializer
		{
			setOpaque(false);
		}

		/*
		 * NOTE: We can't paint the magnifier in the 'glass pane' itself because
		 * it would paint in the same image buffer the other (magnified)
		 * components paint in.
		 */
	};

	@Override
	public void paint(Graphics g) {
		int width = getWidth(), height = getHeight();
		if (paintBuffer == null || paintBuffer.getWidth() != width || paintBuffer.getHeight() != height) {
			// No paint buffer created or component's size changed, (re-)create
			paintBuffer = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		}

		// Paint components into image buffer, then paint the buffer to the
		// on-screen graphics object
		Graphics gimage = paintBuffer.createGraphics();
		gimage.setClip(g.getClip());
		super.paint(gimage);
		g.drawImage(paintBuffer, 0, 0, null);

		Rectangle clip = g.getClipBounds();
		if (magnifierBounds != null && magnifiedArea != null) {
			// Magnification currently active
			
			if (clip.intersects(magnifiedArea) && !clip.contains(magnifierBounds)) {
				// Magnified area may have been modified, but the clipping area
				// does not contain the entire magnifier, initiate repaint
				repaint(magnifierBounds);
			}
			
			if (clip.intersects(magnifierBounds)) {
				Rectangle m = magnifierBounds, s = magnifiedArea;
				// Fill with background color (for when part of the magnified
				// area is outside the component)
				g.setColor(getBackground());
				g.fillRect(m.x, m.y, m.width, m.height);
				// Draw the magnified part of the image
				g.drawImage(paintBuffer, m.x, m.y, m.x + m.width, m.y + m.height,
						s.x, s.y, s.x + s.width, s.y + s.height, null);
				// Draw border (within the magnifier bounds)
				g.setColor(getForeground());
				g.drawRect(m.x, m.y, m.width - 1, m.height - 1);
			}
		}
	}

	/**
	 * Set the center of the magnified area (i.e., the mouse pointer).
	 */
	private void setMagnificationCenter(Point center) {
		if (!magnificationEnabled) {
			magnifiedCenter = null;
			magnifiedArea = null;
			magnifierBounds = null;
			return;
		}
		
		if (magnifiedCenter != null && center.equals(magnifiedCenter)) {
			// Point has not changed
			return;
		}

		magnifiedCenter = center;
		if (magnifiedArea == null) {
			magnifiedArea = new Rectangle(magnifierSize.width / magnificationFactor,
					magnifierSize.height / magnificationFactor);
		}
		magnifiedArea.setLocation(center.x - magnifiedArea.width / 2,
				center.y - magnifiedArea.height / 2);

		Rectangle original = null;
		if (magnifierBounds != null) {
			// Copy previous location of magnifier for below
			original = magnifierBounds.getBounds();
		}

		if (magnifierBounds == null) {
			magnifierBounds = new Rectangle(magnifierCornerGap, magnifierSize);
		}
		// Put magnifier in the opposite corner of the magnified area
		int width = getWidth(), height = getHeight();
		if (center.x < width / 3) {
			// put on the right
			magnifierBounds.x = width - magnifierCornerGap.x - magnifierBounds.width;
		} else if (center.x > 2 * width / 3) {
			// put on the left
			magnifierBounds.x = magnifierCornerGap.x;
		}
		// put on the bottom
		magnifierBounds.y = height - magnifierCornerGap.y - magnifierBounds.height;

		if (original != null && !original.equals(magnifierBounds)) {
			// Position of magnifier changed, repaint previously occupied area
			repaint(original);
		}
		// Repaint (current) magnifier area
		repaint(magnifierBounds);
	}

	/**
	 * Mouse listener for the glass pane that adjusts the magnification view on
	 * mouse movements and dispatches all events to the underlying components in
	 * the content pane.
	 */
	private MouseAdapter mouseListener = new MouseAdapter() {

		@Override
		public void mouseMoved(MouseEvent e) {
			handleEvent(e);
		}

		@Override
		public void mouseDragged(MouseEvent e) {
			handleEvent(e);
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			handleEvent(e);
		}

		@Override
		public void mousePressed(MouseEvent e) {
			handleEvent(e);
		}

		public void mouseExited(MouseEvent e) {
			redispatchEvent(e);
			if (magnifierBounds != null) {
				// Hide magnifier, repaint the area it occupied
				Rectangle bounds = magnifierBounds;
				magnifierBounds = null;
				repaint(bounds);
			}
		}

		@Override
		public void mouseEntered(MouseEvent e) {
			handleEvent(e);
		}

		@Override
		public void mouseClicked(MouseEvent e) {
			handleEvent(e);
		}

		public void mouseWheelMoved(MouseWheelEvent e) {
			handleEvent(e);
		}
		
		private void handleEvent(MouseEvent e) {
			redispatchEvent(e);
			// Set magnification center *after* the event, so any changes in the
			// user interface (such as a list item being selected) are already
			// reflected in any triggered repaints
			setMagnificationCenter(e.getPoint());
		}

		private void redispatchEvent(MouseEvent e) {
			// Dispatch the event to the component in the content pane
			if (e.getSource() == glassPane) {
				Point p = e.getPoint();
				// NOTE: Glass pane and content pane have same bounds, no need
				// to convert the point from the event here
				Component target = SwingUtilities.getDeepestComponentAt(contentPane, p.x, p.y);
				if (target != null) {
					MouseEvent forwardEvent = SwingUtilities
							.convertMouseEvent(glassPane, e, target);
					target.dispatchEvent(forwardEvent);
				}
			}
		}
	};
}
