package gcspy.vis;

import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.DefaultListSelectionModel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;

/**
 * Manages markers in a {@link JList} by registering itself for mouse events and
 * automatically adding ranges marked or unmarked by the user using dragging.
 * Expects the list's cell renderer to call {@link #isMarked(int)} and render
 * marked cells different from non-marked cells. Calls
 * {@link JList#repaint(Rectangle)} on the list with the regions affected by
 * marking in progress or after changes.
 * 
 * @see ListCellRenderer
 * @see JList#repaint(Rectangle)
 */
class MarkerManager {
	/**
	 * Default mouse button for marking items.
	 * 
	 * @see MouseEvent
	 */
	public static final int DEFAULT_MOUSE_BUTTON = MouseEvent.BUTTON3;

	private ListSelectionModel markers;

	private final JList list;

	private MouseAdapter mouseAdapter = new MarkersMouseAdapter();
	private int markButton = DEFAULT_MOUSE_BUTTON;
	private int selBeginIndex = -1, selEndIndex = -1;
	private boolean deselectMode = false;

	/**
	 * Constructor.
	 * 
	 * @param list
	 *            {@link JList} instance this object should operate on.
	 */
	public MarkerManager(JList list) {
		if (list == null)
			throw new NullPointerException("list");

		this.list = list;

		markers = new DefaultListSelectionModel();
		markers.setSelectionMode(DefaultListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

		list.addMouseListener(mouseAdapter);
		list.addMouseMotionListener(mouseAdapter);
	}

	/**
	 * Returns <code>true</code> if and only if the item with the specified
	 * index is marked. Changes of a marking in progress are already included.
	 * 
	 * @param index
	 *            Index of the specified item.
	 * @return <code>true</code> if the item is marked, <code>false</code>
	 *         otherwise.
	 */
	public boolean isMarked(int index) {
		if (selBeginIndex != -1 && selEndIndex != -1) {
			// Selection is being updated
			int first = selBeginIndex, last = selEndIndex;
			if (first > last) {
				first = selEndIndex;
				last = selBeginIndex;
			}
			if (first <= index && index <= last) {
				return (!deselectMode);
			}
		}

		return markers.isSelectedIndex(index);
	}

	/**
	 * Clears all markers and initiates a repaint of the list.
	 */
	public void clear() {
		if (!markers.isSelectionEmpty()) {
			markers.clearSelection();
			list.repaint();
		}
	}

	private class MarkersMouseAdapter extends MouseAdapter {
		@Override
		public void mousePressed(MouseEvent e) {
			if (!list.isEnabled())
				return;

			if (e.getSource() == list && e.getButton() == markButton) {
				selBeginIndex = list.locationToIndex(e.getPoint());

				if (selBeginIndex == -1)
					throw new IndexOutOfBoundsException("list location to index: -1");

				// Unmark the dragged range if the initial tile is marked
				deselectMode = markers.isSelectedIndex(selBeginIndex);
			}
		}

		@Override
		public void mouseDragged(MouseEvent e) {
			if (!list.isEnabled())
				return;

			if (e.getSource() == list && selBeginIndex != -1) {
				int prevEndIndex = selEndIndex;

				selEndIndex = list.locationToIndex(e.getPoint());
				if (selEndIndex == -1)
					throw new IndexOutOfBoundsException("list location to index: -1");

				// Only repaint affected cells
				Rectangle rect = list.getCellBounds(selBeginIndex, selEndIndex);
				if (prevEndIndex != -1) {
					rect = rect.union(list.getCellBounds(selBeginIndex, prevEndIndex));
				}
				list.repaint(rect);
			}
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			if (!list.isEnabled())
				return;

			if (e.getSource() == list && selBeginIndex != -1 && e.getButton() == markButton) {

				selEndIndex = list.locationToIndex(e.getPoint());
				if (selEndIndex == -1)
					throw new IndexOutOfBoundsException("list location to index: -1");

				if (deselectMode) {
					markers.removeSelectionInterval(selBeginIndex, selEndIndex);
				} else {
					markers.addSelectionInterval(selBeginIndex, selEndIndex);
				}

				selBeginIndex = selEndIndex = -1;
				deselectMode = false;

				list.repaint();
			}
		}
	}
}