package gcspy.vis;

import java.awt.Color;

/**
 * Item for display in a {@link PropertyListView}.
 * 
 * @author Peter Hofer
 */
public class PropertyItem {

	private final String name;
	private final boolean isHeading;

	private final String value;

	private final boolean hasPercentage;
	private final double percentage;
	private final Color color;

	/**
	 * Constructor for an item that consists only of a name and no value and
	 * that is optionally a heading.
	 * 
	 * @param name
	 *            Short description of this summary item.
	 * @param isHeading
	 *            <code>true</code> if this item is a heading (typically for the
	 *            succeeding items in the view).
	 */
	public PropertyItem(String name, boolean isHeading) {
		this(name, isHeading, "", false, -1, null);
	}

	/**
	 * Constructor for items consisting of name and value.
	 * 
	 * @param name
	 *            Short description of this summary item.
	 * @param value
	 *            Display value of this item.
	 */
	public PropertyItem(String name, String value) {
		this(name, false, value, false, -1, null);
	}

	/**
	 * Constructor for items consisting of name and value and displayed with a
	 * bar according to a percentage.
	 * 
	 * @param name
	 *            Short description of this summary item.
	 * @param value
	 *            Display value of this item.
	 * @param percentage
	 *            Percentage to show in the bar (<i>&gt;= 0</i> and <i>&lt;=
	 *            100</i>).
	 * @param color
	 *            The color of the bar.
	 */
	public PropertyItem(String name, String value, double percentage, Color color) {
		this(name, false, value, true, percentage, color);
	}

	/**
	 * Constructor for all possible combinations.
	 * 
	 * @param name
	 *            Short description of this summary item.
	 * @param isHeading
	 *            <code>true</code> if this item is a heading (typically for the
	 *            succeeding items in the view).
	 * @param value
	 *            Display value of this item.
	 * @param hasPercentage
	 *            <code>true</code> if this item has a percentage to display in
	 *            the bar.
	 * @param percentage
	 *            Percentage to show in the bar (<i>&gt;= 0</i> and <i>&lt;=
	 *            100</i>). May be an arbitrary value if has
	 *            <code>hasPercentage</code> is <code>false</code>.
	 * @param color
	 *            The color of the bar (may be <code>null</code>).
	 */
	public PropertyItem(String name, boolean isHeading, String value, boolean hasPercentage, double percentage,
			Color color) {
		if (hasPercentage && (percentage < 0 || percentage > 100))
			throw new IllegalArgumentException("Not a valid percentage: " + percentage);

		this.name = name;
		this.value = value;
		this.isHeading = isHeading;
		this.hasPercentage = hasPercentage;
		this.percentage = percentage;
		this.color = color;
	}

	public String getName() {
		return name;
	}

	public String getValue() {
		return value;
	}

	public boolean isHeading() {
		return isHeading;
	}

	public boolean hasPercentage() {
		return hasPercentage;
	}

	public double getPercentage() {
		return percentage;
	}

	public Color getColor() {
		return color;
	}
}
