/**
 ** $Id: SpaceManager.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.client.SpaceCommandListener;
import gcspy.vis.plugins.PluginFrame;
import gcspy.vis.plugins.PluginListener;
import gcspy.vis.plugins.PluginManagerInterface;
import gcspy.vis.utils.Factory;
import gcspy.vis.utils.IconLoader;

import java.awt.BorderLayout;
import java.awt.Window;
import java.awt.Dialog.ModalityType;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import net.miginfocom.swing.MigLayout;
import bibliothek.gui.dock.common.DefaultSingleCDockable;
import bibliothek.gui.dock.common.action.CAction;
import bibliothek.gui.dock.common.action.CButton;
import bibliothek.gui.dock.common.intern.CDockable;
import bibliothek.gui.dock.common.intern.DefaultCDockable;

/**
 * Visual component for a space
 * 
 * @author Tony Printezis
 * @author <a href="http://www/cs/kent.ac.uk">Richard Jones</a>
 * @author Peter Hofer
 */
public class SpaceManager extends JComponent implements DisconnectListener {

	private JComboBox streamChooser;
	private Object multipleStreamsItem;
	
	private TileManager tileManager;
	
	private TileStyle colorConfig;
	
	private JToolBar toolBar;

	private Space space;

	private int[] selectedStreams = { 0 };

	private JToggleButton summaryB;
	private JToggleButton legendB;
	private JButton colorsB;

	private JToggleButton magnificationB; 
	
	private ZoomComponent zoomSlider;

	private InterpreterEventProvider interpreter;

	private Window owner;

	private List<PluginFrame> pluginFrames = new ArrayList<PluginFrame>();

	private TilePropertyView tilePropertyView;

	private LegendView legendView;
	private DefaultCDockable legendDockable;

	private SpaceSummary summaryView;
	private DefaultCDockable summaryDockable;

	/*
	 * Listeners
	 */

	private ChangeListener tileSizeListener = new ChangeListener() {
		@Override
		public void stateChanged(ChangeEvent e) {
			ZoomComponent source = (ZoomComponent) e.getSource();
			tileManager.setTileSize(TileManager.TILE_SIZES[source.getValue()]);
		}
	};
	
	private ActionListener magnificationListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			tileManager.setMagnificationEnabled(magnificationB.isSelected());
		}
	};
	
	private SpaceCommandListener spaceListener = new SpaceCommandListener() {

		@Override
		public void onSpaceCmd(Space s) {
			if (space.getID() == s.getID()) {
				space = s;
				updateTileProperties();
			}
		}
	};

	private ActionListener streamSelectionListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			int index = streamChooser.getSelectedIndex();
			if (index != -1) {
				int[] streams;
				if (streamChooser.getSelectedItem() == multipleStreamsItem) {
					// Let user choose one or more streams to display
					int nstreams = space.getStreamNum();
					Box panel = Box.createVerticalBox();
					panel.add(new JLabel("Please select the streams to display:"));
					JCheckBox[] options = new JCheckBox[nstreams];
					for (int i = 0; i < nstreams; ++i) {
						options[i] = new JCheckBox(space.getStream(i).getName());
						panel.add(options[i]);
					}
					// Preselect currently shown streams
					for (int streamId : selectedStreams) {
						options[streamId].setSelected(true);
					}
					if (JOptionPane.showConfirmDialog(owner, panel, "Stream Selection",
							JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE) == JOptionPane.OK_OPTION) {
						// User clicked "Ok": Determine indices of selected streams
						int[] selection = new int[options.length];
						int nselected = 0;
						for (int i = 0; i < options.length; ++i) {
							if (options[i].isSelected()) {
								selection[nselected] = i;
								nselected++;
							}
						}
						if (nselected >= 1 && nselected <= 3) {
							streams = Arrays.copyOf(selection, nselected);
						} else {
							// None or too many streams selected: don't change
							// displayed streams
							streams = selectedStreams;
							JOptionPane.showMessageDialog(owner,
									"Please choose at least one and at most three streams to display at once.",
									"Stream Selection", JOptionPane.INFORMATION_MESSAGE);
						}
					} else {
						// User canceled: don't change displayed streams
						streams = selectedStreams;
					}
				} else {
					streams = new int[] { index };
				}

				if (!Arrays.equals(streams, selectedStreams)) {
					// Selected streams changed: update tile view and legend
					selectedStreams = streams;
					tileManager.setStreams(streams);
					legendView.setDisplayedStreams(streams);
				}

				if (streams.length == 1 && index != streams[0]) {
					// If only a single stream was chosen in the "multiple
					// streams" dialog, select its item in the combobox.
					streamChooser.setSelectedIndex(streams[0]);
				}
			}
		}
	};

	private Space getSpace() {
		return space;
	}

	public void disconnect(boolean reconnecting) {
		for (PluginFrame frame : pluginFrames)
			frame.disconnect(reconnecting);
		if (reconnecting)
			pluginFrames.clear();
	}

	private class SummaryListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			summaryDockable.setVisible(!summaryDockable.isVisible());
		}
	}

	private class LegendListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			legendDockable.setVisible(!legendDockable.isVisible());
		}
	}

	private class ColorsListener implements ActionListener, TileStyleListener {
		public void actionPerformed(ActionEvent event) {
			TileStyleSettings frame = new TileStyleSettings(owner, "Display Style",
					colorConfig, space, selectedStreams);
			frame.addTileColorsListener(this);
			frame.setModalityType(ModalityType.APPLICATION_MODAL);
			frame.setVisible(true);
			frame.dispose();
		}
		
		@Override
		public void tileStyleChanged(Object source, TileStyle colors) {
			tileManager.update();
			legendView.update();
		}
	}

	private class ClearMarkersListener implements ActionListener {
		public void actionPerformed(ActionEvent event) {
			tileManager.clearMarkers();
		}
	}

	/*
	 * Tile Manager listener
	 */
	
	private TileSelectionListener tileListener = new TileSelectionListener() {
		@Override
		public void tileSelectionChanged(Object source, int index) {
			updateTileProperties();
		}
	};
	
	/*
	 * Property support
	 */

	private PropertySourceListener<PropertyItem[]> propertySourceListener = new PropertySourceListener<PropertyItem[]>() {
		@Override
		public void propertySourceChanged(PropertyView<PropertyItem[]> view, Object newProvider) {
			if (newProvider != SpaceManager.this) {
				tileManager.clearSelection();
			}
		}
	};

	public void setPropertyView(TilePropertyView view) {
		this.tilePropertyView = view;
		view.addPropertySourceListener(propertySourceListener);
	}

	private void updateTileProperties() {
		int selected = tileManager.getSelectedTile();
		if (selected != -1) {
			PropertyItem[] props = SpaceSummaryGenerator.getTileSummary(space, selected);
			tilePropertyView.setProperties(this, props);
		}
	}

	/*
	 * GUI related stuff
	 */

	private void setupMainPanel(PluginManagerInterface plugins) {
		setLayout(new BorderLayout());

		tileManager = new TileManager(interpreter, getSpace(), selectedStreams, colorConfig);
		tileManager.setMagnificationEnabled(false);
		tileManager.addListener(tileListener);

		add(tileManager, BorderLayout.CENTER);

		toolBar = new JToolBar("Tools and Settings");
		toolBar.setFloatable(false);
		add(toolBar, BorderLayout.NORTH);

		summaryB = new JToggleButton("Summary");
		summaryB.addActionListener(new SummaryListener());
		toolBar.add(summaryB);

		legendB = new JToggleButton("Legend");
		legendB.addActionListener(new LegendListener());
		toolBar.add(legendB);

		colorsB = Factory.createButton("Display Style", new ColorsListener());
		toolBar.add(colorsB);

		// Add plugins
		List<PluginListener> pluginListeners = plugins.getPluginListeners();
		if (pluginListeners.size() > 0) {
			toolBar.addSeparator();

			for (PluginListener listener : pluginListeners) {
				listener.init(owner, pluginFrames, interpreter, space, selectedStreams[0], colorConfig);
				JButton traceB = Factory.createButton(listener.getLabel(), listener);
				toolBar.add(traceB);
			}
		}

		JPanel bottomPanel = new JPanel(new MigLayout("insets 1"));

		streamChooser = new JComboBox();
		int nstreams = space.getStreamNum();
		for (int i = 0; i < nstreams; ++i) {
			streamChooser.addItem(space.getStream(i).getName());
		}
		multipleStreamsItem = "< Show multiple streams... >";
		streamChooser.addItem(multipleStreamsItem);
		streamChooser.addActionListener(streamSelectionListener);
		bottomPanel.add(streamChooser, "push");

		zoomSlider = new ZoomComponent(0, TileManager.TILE_SIZES.length - 1, TileManager.TILE_SIZE_DEFAULT_INDEX);
		zoomSlider.addChangeListener(tileSizeListener);
		bottomPanel.add(zoomSlider, "gap unrelated");

		magnificationB = new JToggleButton(IconLoader.getIcon(IconLoader.MAGNIFIER_ICON),
				tileManager.isMagnificationEnabled());
		magnificationB.addActionListener(magnificationListener);
		bottomPanel.add(magnificationB, "gap unrelated");

		JButton clearMarkersB = Factory.createButton("Clear Markers", new ClearMarkersListener());
		bottomPanel.add(clearMarkersB);

		add(bottomPanel, BorderLayout.SOUTH);
	}

	private void setupDockables() {
		legendView = new LegendView(interpreter, space, selectedStreams, colorConfig);
		legendDockable = new DefaultSingleCDockable("space#" + space.getID() + "_legend",
			"Legend: " + space.getName(), new JScrollPane(legendView));
		legendDockable.setCloseable(true);
		legendDockable.addCDockableStateListener(new ButtonSelectDockableListener(legendB));

		summaryView = new SpaceSummary(owner, interpreter, space);
		CAction copyAction = new CButton("Copy to Clipboard", IconLoader.getIcon(IconLoader.COPY_ICON)) {
			@Override
			protected void action() {
				summaryView.copyToClipboard();
			}
		};
		summaryDockable = new DefaultSingleCDockable("space#" + space.getID() + "_summary",
			"Summary: " + space.getName(), summaryView, copyAction);
		summaryDockable.setCloseable(true);
		summaryDockable.addCDockableStateListener(new ButtonSelectDockableListener(summaryB));
	}

	public CDockable getLegendDockable() {
		return legendDockable;
	}

	public CDockable getSummaryDockable() {
		return summaryDockable;
	}

	/*
	 * Constructor
	 */

	/**
	 * Create a new space manager
	 * 
	 * @param owner
	 *            The parent frame
	 * @param interpreter
	 *            The client interpreter
	 * @param space
	 *            the space to manage
	 * @param pluginInterface
	 *            Provides access to plugins
	 * @param active
	 *            True if this space is active
	 */
	public SpaceManager(Window owner, InterpreterEventProvider interpreter,
			Space space, PluginManagerInterface pluginInterface) {

		interpreter.addSpaceListener(spaceListener);

		this.space = space;
		this.owner = owner;
		this.interpreter = interpreter;

		/* FIXME: TileManagerColors does not listen to SPACE commands and never reinitializes if the colors change */
		this.colorConfig = new TileStyle(space);

		setupMainPanel(pluginInterface);
		setupDockables();
		streamChooser.setSelectedIndex(selectedStreams[0]);
	}
}
