package gcspy.vis;

import gcspy.interpreter.DataAccessor;
import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.client.EventCommandListener;
import gcspy.interpreter.client.SpaceCommandListener;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.AbstractListModel;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.plaf.ListUI;

/**
 * View for a space and its tiles using {@link JList}.
 * 
 * @author Peter Hofer
 */
public class TileManager extends JComponent {

	public static final Dimension TILE_SIZE_HUGE = new Dimension(22, 31);
	public static final Dimension TILE_SIZE_LARGE = new Dimension(16, 22);
	public static final Dimension TILE_SIZE_MEDIUM = new Dimension(13, 18);
	public static final Dimension TILE_SIZE_SMALL = new Dimension(11, 15);
	public static final Dimension TILE_SIZE_TINY = new Dimension(9, 13);
	
	/**
	 * Tile sizes, in order from smallest to largest.
	 */
	public static final Dimension[] TILE_SIZES = new Dimension[] { TileManager.TILE_SIZE_TINY,
			TileManager.TILE_SIZE_SMALL, TileManager.TILE_SIZE_MEDIUM, TileManager.TILE_SIZE_LARGE,
			TileManager.TILE_SIZE_HUGE };
	public static final int TILE_SIZE_DEFAULT_INDEX = 2;
	public static final Dimension TILE_SIZE_DEFAULT = TILE_SIZES[TILE_SIZE_DEFAULT_INDEX];

	private Space space;
	private int[] streams;
	
	private TileStyle tileStyle;
	
	private List<TileSelectionListener> listeners = new ArrayList<TileSelectionListener>();

	private JList list;
	private SpaceListModel model;
	private TileRenderer renderer;

	private MarkerManager markers;

	private JScrollPane scrollPane;
	
	private MagnifyingPanel panel;
	
	private InterpreterListener interpreterListener = new InterpreterListener();

	/**
	 * Listener for key events that implements better navigation behavior
	 * compared to the default handling in {@link JList} and various
	 * {@link ListUI} implementations.
	 */
	private KeyListener navigationKeyListener = new KeyListener() {
		/**
		 * Finds the first item in the same row as the item at the given index.
		 */
		private int getFirstInRow(int index) {
			Point point = list.indexToLocation(index);
			point.x = 0;
			return list.locationToIndex(point);
		}

		/**
		 * Finds the last item in the same row as the item at the given index.
		 */
		private int getLastInRow(int index) {
			Point point = list.indexToLocation(index);
			point.x = list.getWidth();
			return list.locationToIndex(point);
		}

		private boolean isHandled(KeyEvent event) {
			int code = event.getKeyCode();
			return (code == KeyEvent.VK_HOME || code == KeyEvent.VK_END || code == KeyEvent.VK_LEFT || code == KeyEvent.VK_RIGHT);
		}

		public void keyTyped(KeyEvent e) {
			if (isHandled(e)) {
				e.consume();
			}
		}

		public void keyPressed(KeyEvent e) {
			if (isHandled(e)) {
				e.consume();
			}
		}

		public void keyReleased(KeyEvent e) {
			int code = e.getKeyCode();
			boolean ctrl = ((e.getModifiersEx() & KeyEvent.CTRL_DOWN_MASK) != 0);
			int selection = list.getSelectedIndex();

			int newSelection = -1;
			switch (code) {
			case KeyEvent.VK_HOME:
				if (ctrl) {
					newSelection = 0;
				} else {
					// Jump to first in row
					newSelection = getFirstInRow(selection);
				}
				break;

			case KeyEvent.VK_END:
				if (ctrl) {
					newSelection = model.getSize() - 1;
				} else {
					// Jump to last in row
					newSelection = getLastInRow(selection);
				}
				break;

			case KeyEvent.VK_LEFT:
				if (ctrl) {
					// Skip multiple cells
					newSelection = (selection - 5);
				} else {
					newSelection = selection - 1;
				}
				break;

			case KeyEvent.VK_RIGHT:
				if (ctrl) {
					// Skip multiple cells
					newSelection = (selection + 5);
				} else {
					newSelection = selection + 1;
				}
				break;
			}
			if (newSelection != -1) {
				list.setSelectedIndex(newSelection);
				list.ensureIndexIsVisible(newSelection);
				e.consume();
			}
		};
	};

	/**
	 * Listener for events from the interpreter that affect the enclosing
	 * {@link TileManager} instance.
	 */
	private class InterpreterListener implements SpaceCommandListener, EventCommandListener {
		@Override
		public void onSpaceCmd(Space arg) {
			if (space == null || space.getID() == arg.getID()) {
				space = arg;
				model.setSpace(arg);
				repaint();
			}
		}

		@Override
		public void onEventCmd(int eventID, int elapsedTime, int compensationTime) {
			repaint();
		}
	}

	/**
	 * {@link ListCellRenderer} that uses {@link TileComponent} to visualize
	 * tiles within a JList.
	 */
	private static class TileRenderer implements ListCellRenderer {
		private final TileComponent component;
		private MarkerManager markers;

		public TileRenderer(TileStyle style) {
			this.component = new TileComponent(style);
		}

		public void setMarkerManager(MarkerManager markers) {
			this.markers = markers;
		}

		@Override
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
				boolean cellHasFocus) {

			component.setTile((Tile) value);

			component.setComponentOrientation(list.getComponentOrientation());
			component.setEnabled(list.isEnabled());

			component.setSelected(isSelected);
			component.setMarked(markers != null && markers.isMarked(index));

			return component;
		}
	}
	
	/**
	 * Model for a {@link JList} backed by a {@link Space} and {@link Stream}.
	 */
	private static class SpaceListModel extends AbstractListModel {

		private Space space;
		private int[] streams;

		void setSpace(Space space) {
			this.space = space;
		}

		void setStreams(int[] streams) {
			this.streams = streams;
		}

		@Override
		public Object getElementAt(int index) {
			if (index >= space.getTileNum())
				throw new IndexOutOfBoundsException();

			for (int stream : streams) {
				DataAccessor accessor = space.getStream(stream).getAccessor();
				if (!accessor.isDataAvailable()) {
					return null;
				}
			}

			return new Tile(space, streams, index);
		}

		@Override
		public int getSize() {
			return space.getTileNum();
		}
	}

	/**
	 * Constructor.
	 * 
	 * @param tileStyle
	 *            A style configuration for the tiles
	 */
	public TileManager(InterpreterEventProvider interpreter, Space space, int[] streams, TileStyle tileStyle) {
		if (space == null)
			throw new NullPointerException("space");
		if (tileStyle == null)
			throw new NullPointerException("tileStyle");
		
		interpreter.addEventListener(interpreterListener);
		interpreter.addSpaceListener(interpreterListener);
		
		this.space = space;
		this.streams = streams;
		this.tileStyle = tileStyle;
		
		model = new SpaceListModel();
		model.setSpace(space);
		model.setStreams(streams);
		
		renderer = new TileRenderer(tileStyle);
		
		list = new JList(model);
		list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		list.setCellRenderer(renderer);
		list.setLayoutOrientation(JList.HORIZONTAL_WRAP);
		list.setVisibleRowCount(-1);
		list.addListSelectionListener(selectionListener);
		list.addKeyListener(navigationKeyListener);
		
		setTileSize(TILE_SIZE_DEFAULT);
		
		markers = new MarkerManager(list);
		renderer.setMarkerManager(markers);

		scrollPane = new JScrollPane(list);
		scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		
		panel = new MagnifyingPanel();
		panel.setLayout(new BorderLayout());
		panel.add(scrollPane);
		
		setLayout(new BorderLayout());
		add(panel, BorderLayout.CENTER);
	}
	
	public void setTileSize(Dimension d)
	{
		setTileSize(d.width, d.height);
	}
	
	public void setTileSize(int width, int height)
	{
		list.setFixedCellWidth(width);
		list.setFixedCellHeight(height);
	}
	
	@Override
	public void setEnabled(boolean enabled) {
		list.setEnabled(enabled);
	}
	
	public void setMagnificationEnabled(boolean enabled) {
		panel.setMagnificationEnabled(enabled);
	}

	public boolean isMagnificationEnabled() {
		return panel.isMagnificationEnabled();
	}

	public Space getSpace() {
		return space;
	}
	
	public int[] getStreams()
	{
		return streams;
	}
	
	public void setStreams(int[] streams)
	{
		if (!Arrays.equals(streams, this.streams)) {
			this.streams = streams;
			model.setStreams(streams);
			update();
		}
	}
	
	public int getSelectedTile() {
		return list.getSelectedIndex();
	}

	/**
	 * Clears the current selection.
	 */
	public void clearSelection() {
		list.clearSelection();
	}

	public void update() {
		list.setBackground(tileStyle.getBackgroundColor());
		repaint();
	}

	public void clearMarkers() {
		markers.clear();
	}
	/*
	 * Tile manager listeners
	 */

	public void addListener(TileSelectionListener listener) {
		listeners.add(listener);
	}

	private ListSelectionListener selectionListener = new ListSelectionListener() {
		@Override
		public void valueChanged(ListSelectionEvent e) {
			if (e.getSource() == list && !e.getValueIsAdjusting()) {
				fireSelectionChanged(list.getSelectedIndex());
			}
		}
	};

	private void fireSelectionChanged(int index) {
		for (TileSelectionListener listener : listeners)
			listener.tileSelectionChanged(this, index);
	}
}
